<?php

interface READERInterface {
    public function __construct(&$props);

    public function GetDatabaseList($flags = 0);
    public function GetGroupList($flags = 0);
    public function GetGroupInfo(LOGGROUP &$grp = NULL, $flags = 0);
    public function GetItemList(LOGGROUP &$grp = NULL, MASK &$mask = NULL, $flags = 0);
    public function GetExperimentList($flags = 0);
    public function GetMaskList(LOGGROUP &$grp = NULL, $flags = 0);
    
    public function CreateGroup(array &$ginfo = NULL, $flags = 0);
    public function CreateInterval(LOGGROUP &$group = NULL, array &$iinfo = NULL, $flags = 0);
    public function CreateMask(LOGGROUP &$grp = NULL, array &$minfo = NULL, $flags = 0);
    
    public function GetGroups();
    public function GetGroupSize(LOGGROUP &$grp = NULL);
    
    public function ImportUnixTime($unix_time);
    public function ExportUnixTime($db_time);

    public function HaveData(LOGGROUP &$grp = NULL, $from = 0, $to = 0);
    public function GetData(LOGGROUP &$grp = NULL, $from = 0, $to = 0);
    public function PushData(LOGGROUP &$grp, $time, $data);
    public function RemoveData(LOGGROUP &$grp, $time);
    
    public function Backup(&$binfo);
}

abstract class READER implements READERInterface {
 var $req;
 var $server;
 var $opts;

 var $srvid, $dbname;
 var $start_date;
 var $end_date;

 var $time_format;
 var $time_zone;
 var $time_module;

 var $gmt_timezone;
 
 var $group_class;
 
 var $groups;			// Per group options, configs, etc...
 
 const LIST_ALL = 1;
 const NEED_INFO = 2;
 const NEED_COUNT = 4;
 const NEED_ITEMINFO = 8;

 const BACKUP_FULL = 1;

 const PARAM_MISSING_ITEMS = 1;	// Parameters related to recovery of missing/extra items
 
 function __construct(&$props) {
    if ($props instanceof REQUEST) {
	$this->req = &$props;
	$this->server = $this->req->GetServerInfo();
	
	$this->srvid = $props->props['db_server'];
    } else {
	$this->req = false;
	$this->server = &$props;

	$this->srvid = false;
    }
    
    if (isset($this->server['database']))
	$this->dbname = $this->server['database'];
    else 
	$this->dbname = false;
	
    $this->time_format = "U.u";
    $this->time_zone = new DateTimeZone("GMT");

    $this->gmt_timezone = new DateTimeZone("GMT");

    $this->opts = new OPTIONS($props);

    $date_limit = $this->opts->Get('date_limit');
    if ($date_limit) {
	if (is_array($date_limit)) {
	    if (is_int($date_limit[0])) $this->start_date = $date_limit[0];
	    else $this->start_date = strtotime($date_limit[0]);
	    if (is_int($date_limit[1])) $this->end_date = $date_limit[1];
	    else $this->end_date = strtotime($date_limit[1]);
	} else {
	    if (is_int($date_limit)) $this->start_date = $date_limit;
	    else $this->start_date = strtotime($date_limit);
	    $this->end_date = false;
	}
    } else {
	$this->start_date = false;
	$this->end_date = false;
    }

    $time_module = $this->opts->Get('time_module');
    if ($time_module) {
	if (!include_once("time/" . strtolower($time_module) . '.php')) {
	    throw new ADEIException(translate("Unsupported time module is configured: \"%s\"", $time_module));
	}
	
	$time_options = $this->opts->Get('time_options');
	$cl = strtoupper($time_module);
	$this->time_module = new $cl($this, $time_options);
    } else $this->time_module = false;
    
    $this->group_class = false;
    $this->groups = array();
 }

 function ImportTime(DateTime $dt) {
    if ($this->time_module) return $this->time_module->ImportTime($dt);
    
    $dt->setTimezone($this->time_zone);
    return $dt->format($this->time_format);
 }

 function ExportTime($db_time) {
    if ($this->time_module) return $this->time_module->ExportTime($db_time);

    if ($this->time_zone) {
        $ctz = date_default_timezone_get();
	date_default_timezone_set($this->time_zone->getName());
	$dt = new DateTime($db_time);
	date_default_timezone_set($ctz);

//	$dt->setTimezone($this->gmt_timezone);
	return $dt;
    }
    return new DateTime($db_time);
 }

 function ImportUnixTime($unix_time) {
    $itime = (int)floor($unix_time);

    if ($itime==$unix_time) {
            // Time zone should be specified, esle "e" output is invalid
	return $this->ImportTime(new DateTime("@$itime", $this->gmt_timezone));
    } else {
	if (is_float($unix_time)) $subsec = strchr(sprintf("%F", $unix_time), '.');
	else $subsec = strchr($unix_time, '.');
	
	    // DS: Due to the bug in PHP (precision is limited to 10ns)
	if (strlen($subsec) > 9) $subsec = substr($subsec, 0, 9);
	return $this->ImportTime(new DateTime(strftime("%Y-%m-%dT%H:%M:%S", $itime) . $subsec, $this->gmt_timezone));
//	return $this->ImportTime(new DateTime(strftime("%Y/%m/%d %H:%M:%S", $itime) . $subsec, $this->gmt_timezone));
    }
 }
 
 function ExportUnixTime($db_time) {
    $dt = $this->ExportTime($db_time);
    return $dt->format("U.u");
 }

 function GetDatabaseList($flags = 0) {
    if ($this->req)
	$server = $this->req->GetServerConfig();
    else
	throw new ADEIException(translate("The data source server is not specified"));

    $dblist = array();
    foreach ($server['database'] as $name) {
	$dblist[$name] = array(
	    'name' => $name
	);
    }
    
    return $dblist;
 }

 function GetGroupList($flags = 0) {
    return $this->GetGroupInfo($gr=NULL, $flags);
 }

 function GetExperimentList($flags = 0) {
    return array();
 }
 
 function GetMaskList(LOGGROUP &$grp = NULL, $flags = 0) {
    return array();
 }

 function CheckGroup(LOGGROUP &$grp = NULL) {
    if ($grp) {
	if (($this->group_class)&&(!$grp instanceof $this->group_class))
	    throw Exception(translate("Invalid LOGGROUP supplied"));
    } else {
	$grinfo = $this->req->GetGroupInfo();
	$grp = $this->CreateGroup($grinfo);
    }
    return $grp;
 }

 function CreateGroup(array &$ginfo = NULL, $flags = 0) {
    if (!$ginfo) $ginfo = $this->req->GetGroupInfo();

    if ($this->group_class) return new $this->group_class($ginfo, $this, $flags);
    return new LOGGROUP($ginfo, $this, $flags);
 }

 function CreateMask(LOGGROUP &$grp = NULL, array &$minfo = NULL, $flags = 0) {
    if (!$minfo) {
	if ($this->req instanceof GROUPRequest)
	    $minfo = $this->req->GetMaskInfo();
    }
    //$grp = $this->CheckGroup($grp);
    
    return new MASK($minfo);
 }

 function CreateInterval(LOGGROUP &$grp = NULL, array &$iinfo = NULL, $flags = 0) {
    if (!$iinfo) {
	if ($this->req instanceof DATARequest)
	    $iinfo = $this->req->GetIntervalInfo();
    }
    $grp = $this->CheckGroup($grp);
    
    $ivl = new INTERVAL($iinfo, $this, $grp, $flags);
    
    if ((($this->start_date)&&($ivl->GetWindowStart() < $this->start_date))||(($this->end_date)&&($ivl->GetWindowEnd() > $this->end_date))) {
	$ivl->Limit($this->start_date, $this->end_date);
    }
    
    return $ivl;
 }


 function GetGroups() {
    $groups = $this->GetGroupList(); 

    $list = array();
    foreach (array_keys($groups) as $group) {
    	$ginfo = array("db_group" => $group);
	array_push($list, $this->CreateGroup($ginfo));
    }

    return $list;
 }

 function GetGroupSize(LOGGROUP &$grp = NULL) {
    return sizeof($this->GetItemList($grp, $mask = new MASK()));
 }
 
 function HaveData(LOGGROUP &$grp = NULL, $from = 0, $to = 0) {
    return true;
 }

 function PushData(LOGGROUP &$grp, $time, $data) {
    throw new ADEIException(get_class($this) . ". PushData is not implemented");
 }
 
 function RemoveData(LOGGROUP &$grp, $time) {
    throw new ADEIException(get_class($this) . ". RemoveData is not implemented");
 }
    
 /* DS: check for invalid columns number */
 function Clean(&$lg, $from = 0, $to = 0) {
    $data = $this->GetData($lg, $from, $to);
    foreach ($data as $t => $value) {
	if (sizeof($value)==0) {
	    $this->RemoveData($lg, $t);
	}
	
	$check = true;
	foreach ($value as $v) {
	    if ($v) {
		$check = false;
		break;
	    }
	}
	
	if ($check) {
	    $this->RemoveData($lg, $t);
	}
    }
 }

 function GetGroupParameters(LOGGROUP $grp, $prm = false) {
    global $DEFAULT_MISSING_VALUE;
    
    if (!$this->groups[$grp->gid]) $this->groups[$grp->gid] = array();
    $group = &$this->groups[$grp->gid];

    switch ($prm) {
	case READER::PARAM_MISSING_ITEMS:
	    if (!isset($group['fix_missing_items'])) {
		$opts = $this->req->GetGroupOptions($grp);
	
		$group['fix_extra_items'] = $opts->Get("fix_extra_items", false);
		$group['fix_missing_items'] = $opts->Get("fix_missing_items", false);
		$group['missing_value'] = $opts->Get("missing_value", $DEFAULT_MISSING_VALUE);
		
		if ($group['fix_missing_items']) {
		    if (is_array($group['fix_missing_items'])) {
			if (!isset($group['fix_missing_items']['limit']))
			    $group['fix_missing_items']['limit'] = 1;
			if (!isset($group['fix_missing_items']['value']))
			    $group['fix_missing_items']['value'] = NULL;
		    } else {
			$group['fix_missing_items'] = array(
			    'limit' => 1,
			    'value' => NULL
			);
		    }
		}
	    }
	break;
    }
    
    return $group;
 }

 function FixRecordItems(array &$row, LOGGROUP $grp, $expected = 0) {
    if (!$expected) {
/*	
	The group size could change during the class execution,
	therefore, we is not allowed to cache result. I think.
	
	if (!$group['size']) $group['size'] = $this->GetGroupSize($grp);
	$expected = $group['size'];
*/
	$expected = $this->GetGroupSize($grp);
    }
 
    if (sizeof($row) == $expected) return true;
    

    $group = $this->GetGroupParameters($grp, READER::PARAM_MISSING_ITEMS);

    if (sizeof($row) > $expected) {
	if ($group['fix_extra_items']) {
	    $keys = array_keys($row);
	    for ($i = sizeof($keys) - 1;$i >= $expected; $i--) {
		unset($row[$keys[$i]]);
	    }
	    return true;
	}
    } else {
	if (($group['fix_missing_items'])&&(sizeof($row) > $group['fix_missing_items']['limit'])) {
    	    $row = array_merge($row, array_fill(0, $expected - sizeof($row), $group['missing_value']));
	    return true;
	}
    }		
    
    return false;
 }

 function Backup(&$binfo) {
    throw new ADEIException(get_class($this) . ". Backup is not implemented");
 }

 function Export(DATAHandler &$h = NULL, LOGGROUP &$grp = NULL, MASK &$mask = NULL, $from = 0, $to = 0, $opts = 0) {
    if (!$h) $h = new CSVHandler();
    $grp = $this->CheckGroup($grp);

    $names = $this->GetItemList($grp, $mask);
    $data = $this->GetData($grp, $from, $to);

    if (($mask)&&(is_array($mask->ids))) {
	$mask_ids = $mask->ids;
	$mask_max_id = max($mask_ids);
	if ($mask_max_id >= $this->GetGroupSize($grp)) {
	    throw new ADEIException(translate("Invalid item mask is supplied. The ID:%d refers non-existing item.", $mask_max_id));
	}
    } else {
	$mask_ids = false;
    }

    $h->Start();
    $h->HeaderStart();
    $h->TimeHeader();
    $columns = 0;
    foreach ($names as $name) {
	$h->DataHeader($name["name"], $columns);
	$columns++;
    }
    $h->HeaderEnd();
    
    foreach ($data as $time => $row) {
	if ($mask_ids) {
	    if (sizeof($row) <= $mask_max_id) {
#		echo sizeof($row) . " (1)\n";
		continue;
	    }
	} else {
	    if (sizeof($row) != $columns) {
#		echo sizeof($row) . " (2)\n";
		if (!$this->FixRecordItems($row, $grp, $columns)) continue;
	    }
	}
	
	$h->VectorStart();
	$h->TimeValue($time);

	if ($mask_ids) {
	    foreach ($mask_ids as $i) {
		if ($row[$i] === NULL) {
		    if (!$group) $group = $this->GetGroupParameters($grp, READER::PARAM_MISSING_ITEMS);
		    $h->DataValue($group['missing_value'], $i);

		} else $h->DataValue($row[$i], $i);
	    }
	} else {
	    foreach ($row as $i => &$column) {
		if ($column === NULL) {
		    if (!$group) $group = $this->GetGroupParameters($grp, READER::PARAM_MISSING_ITEMS);
		    $h->DataValue($group['missing_value'], $i);
		} else $h->DataValue($column, $i);
	    }
	}
	$h->VectorEnd();
    }

    $h->End();
 }

 function ExportCSV(STRINGHandler &$h = NULL, LOGGROUP &$grp = NULL, MASK &$mask = NULL, $from = 0, $to = 0, $opts = 0) {
    return $this->Export(new CSVHandler($h), $grp, $msk, $from, $to, $opts);
 }

/*
 function ExportCSV(STRINGHandler &$h = NULL, LOGGROUP &$grp = NULL, $from = 0, $to = 0, $opts = 0) {
    if (!$h) $h = new STRINGStreamHandler();
    $grp = $this->CheckGroup($grp);

    $subseconds = !$this->opts->Get('ignore_subsecond');

    $names = $this->GetItemList($grp);
    $data = $this->GetData($grp, $from, $to);

    $columns = 0;
    $h->Process("Date/Time");
    foreach ($names as $name) {
	$h->Process(", " . preg_replace("/,/", " ", $name["name"]));
	$columns++;
    }
    $h->Process("\r\n");

    $saved_tz = date_default_timezone_get ();
    date_default_timezone_set("UTC");
    foreach ($data as $time => $row) {
	if (sizeof($row) != $columns) continue;
	
	if ($subseconds) {
	    $subsec = strchr(sprintf("%.6F", $time), '.');
	    $h->Process(date("d-M-y H:i:s", $time) . $subsec);
	} else 
	    $h->Process(date("d-M-y H:i:s", $time));
	foreach ($row as $column) {
	    $h->Process(", " . $column);
	}
	$h->Process("\r\n");
    }
    date_default_timezone_set($saved_tz);
 }
*/
}



class NULLReader implements READERInterface {
    function __construct(&$props) {}

    function GetDatabaseList($flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }

    function GetGroupList($flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetGroupInfo(LOGGROUP &$grp = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetItemList(LOGGROUP &$grp = NULL, MASK &$mask = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetExperimentList($flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetMaskList(LOGGROUP &$grp = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function CreateGroup(array &$ginfo = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function CreateInterval(LOGGROUP &$group = NULL, array &$iinfo = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function CreateMask(LOGGROUP &$grp = NULL, array &$minfo = NULL, $flags = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetGroups() {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetGroupSize(LOGGROUP &$grp = NULL) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function ImportUnixTime($unix_time) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function ExportUnixTime($db_time) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }

    function HaveData(LOGGROUP &$grp = NULL, $from = 0, $to = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function GetData(LOGGROUP &$grp = NULL, $from = 0, $to = 0) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function PushData(LOGGROUP &$grp, $time, $data) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function RemoveData(LOGGROUP &$grp, $time) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
    
    function Backup(&$binfo) {
	throw new ADEIException(get_class($this) . ". " . __METHOD__ . " is not implemented");
    }
}

?>